"""
Helpful utility functions
"""
from __future__ import unicode_literals

import os
import random
import re
from datetime import datetime, timedelta
from dateutil import parser, tz
import logging

logger = logging.getLogger(__name__)

NONCE_V1_PREFIX = '001'
NONCE_V2_PREFIX = '002'

DEFAULT_NONCE_EXPIRY_SECONDS = (1 * 60 * 60)  # one hour

_valid_chars = None


def make_nonce(expiry=None):
    """
    Create a nonce with expiration timestamp included

    :param expiry: a `datetime` that indicates when the nonce self-expires,
        defaults to now + 30 minutes

    :return: nonce
    """
    global _valid_chars

    if not _valid_chars:
        _valid_chars = ''
        # iterate over all the ascii characters for a list of all alpha-numeric characters
        for char_index in range(0, 128):
            if chr(char_index).isalpha() or chr(char_index).isalnum():
                _valid_chars += chr(char_index)

    if not expiry:
        now = datetime.utcnow().replace(tzinfo=tz.tzutc())
        expiry = (now + timedelta(seconds=DEFAULT_NONCE_EXPIRY_SECONDS))

    time_format = '%Y-%m-%dT%H:%M:%SZ'
    time_component = expiry.strftime(time_format)

    random_str = ''
    random_chr = random.SystemRandom()
    for i in range(0, 6):
        random_str += random_chr.choice(_valid_chars)

    return '{prefix}{time_str}{random_str}'.format(
        prefix=NONCE_V2_PREFIX,
        time_str=time_component,
        random_str=random_str,
    )


def _default_nonce_verifier(nonce):
    oneid_directory = os.path.join(os.path.expanduser('~'), '.oneid')
    nonce_cache_fn = os.path.join(oneid_directory, 'used_nonces.txt')

    if not os.path.exists(oneid_directory):
        os.makedirs(oneid_directory)

    if os.path.exists(nonce_cache_fn):
        count = 0
        with open(nonce_cache_fn, 'r') as fd:
            for saved_nonce in fd:
                saved_nonce = saved_nonce.rstrip()
                if saved_nonce == nonce:
                    return False
                count += 1
        if count > 10000:  # pragma: no cover  TODO: mock or attach handler to logger
            logger.warning(
                'nonce cache is getting full (%n entries), consider alternate store',
                count
            )
    return True


_nonce_verifier = _default_nonce_verifier


def _default_nonce_burner(nonce):
    oneid_directory = os.path.join(os.path.expanduser('~'), '.oneid')
    nonce_cache_fn = os.path.join(oneid_directory, 'used_nonces.txt')

    if not os.path.exists(oneid_directory):
        os.makedirs(oneid_directory)

    with open(nonce_cache_fn, 'a+') as fd:
        fd.write(nonce + '\n')

    return True


_nonce_burner = _default_nonce_burner


def set_nonce_handlers(nonce_verifier, nonce_burner):
    """
    Sets the functions to verify nonces and record their use.

    By default, the nonces are saved in a local file
    named `~/.oneid/used_nonces.txt` (or equivalent)

    :param nonce_burner: function to be called to verify. Passed one argument, the nonce
    :param nonce_verifier: function to be called to burn. Passed one argument, the nonce
    """
    global _nonce_burner, _nonce_verifier

    _nonce_verifier = nonce_verifier
    _nonce_burner = nonce_burner


def verify_nonce(nonce, expiry=None):
    """
    Ensure that the nonce is correct, and not from the future

    Callers should also store used nonces and reject messages
    with previously-used ones.

    :param nonce: Nonce as created with :func:`~oneid.nonces.make_nonce`
    :param expiry: If not None, a `datetime` before which the nonce is not valid
    :return: True only if nonce meets validation criteria
    :rtype: bool
    """
    NONCE_REGEX = (
        r'^00[12]'
        r'[2-9][0-9]{3}-(0[1-9]|1[0-2])-(0[1-9]|[12][0-9]|3[01])'
        r'T([01][0-9]|2[0-3])(:[0-5][0-9]){2}Z'
        r'[A-Za-z0-9]{6}$'
    )

    if not re.match(NONCE_REGEX, nonce):
        logger.debug('incorrectly-formatted nonce: %s', nonce)
        return False

    nonce_date = parser.parse(nonce[3:-6])

    if expiry and (nonce_date < expiry):
        logger.debug('out-of-date-range nonce: %s, expiry=%s', nonce, expiry)
        return False

    now = datetime.utcnow().replace(tzinfo=tz.tzutc())

    if nonce[:3] == NONCE_V1_PREFIX:
        now_ish = (now + timedelta(minutes=2))

        if nonce_date > now_ish:
            logger.debug('nonce from the future, invalid: %s', nonce)
            return False

    elif now > nonce_date:
        logger.debug('nonce is expired: %s', nonce)
        return False

    return _nonce_verifier(nonce)


def burn_nonce(nonce):
    return _nonce_burner(nonce)



# -*- coding: utf-8 -*-

"""
Utility functions for dealing with JOSE objects generally.

Mostly used internally, but may be useful for external callers.
"""

import re
import json
import six
import time
import logging

from datetime import datetime
from dateutil import tz

from . import nonces, utils, exceptions

logger = logging.getLogger(__name__)


B64_URLSAFE_RE = '[0-9a-zA-Z-_]+'
COMPACT_JWS_RE = r'^{b64}\.{b64}\.{b64}$'.format(b64=B64_URLSAFE_RE)


TOKEN_EXPIRATION_TIME_SEC = nonces.DEFAULT_NONCE_EXPIRY_SECONDS


def is_compact_jws(msg):
    """
    Determine if a given message is a compact JWS (or JWT) or not.
    Does not necessarily mean that it is valid or authentic.

    :param msg: message to inspect
    :type msg: str
    :returns: True if the message is a compact JWS, False otherwise
    :rtype: bool
    """
    return bool(re.match(COMPACT_JWS_RE, msg))


def is_jws(msg, json_decoder=json.loads):
    """
    Determine if a given message is a JWS or not (compact or otherwise).
    Does not necessarily mean that it is valid or authentic.

    :param msg: message to inspect
    :type msg: str or dict
    :param json_decoder: a function to decode JSON into a :py:class:`dict`. Defaults to `json.loads`
    :returns: True if the message is a JWS, False otherwise
    :rtype: bool
    """
    REQUIRED_FIELDS = ['payload', 'signatures']

    if isinstance(msg, six.string_types):
        if is_compact_jws(msg):
            return True
        msg = json_decoder(msg)

    if not isinstance(msg, dict):
        return False

    if not all([k in msg for k in REQUIRED_FIELDS]):
        return False

    return True


def is_jwe(msg, json_decoder=json.loads):
    """
    Determine if a given message is a JWE or not.
    Does not necessarily mean that it is valid or authentic.

    :param msg: message to inspect
    :type msg: str or dict
    :param json_decoder: a function to decode JSON into a :py:class:`dict`. Defaults to `json.loads`
    :returns: True if the message is a JWS, False otherwise
    :rtype: bool
    """
    REQUIRED_FIELDS = ['iv', 'ciphertext', 'tag', 'recipients']

    if isinstance(msg, six.string_types):
        msg = json_decoder(msg)

    if not isinstance(msg, dict):
        return False

    if not all([k in msg for k in REQUIRED_FIELDS]):
        return False

    return True


def get_jwe_shared_header(jwe, json_decoder=json.loads):
    """
    Extract shared (non-encrypted) header values from a JWE

    :param jwe: JWE to extract header field values from
    :type jwe: str or dict
    :param json_decoder: a function to decode JSON into a :py:class:`dict`. Defaults to `json.loads`
    :returns: header fields and values
    :rtype: dict
    :raises: :py:class:`~oneid.exceptions.InvalidFormatError`: if not a valid JWE
    """
    jwe_dict = as_dict(jwe, json_decoder)

    if not is_jwe(jwe_dict):
        logger.debug('attempt to get header values from a non-jwe (type: %s): %s', type(jwe), jwe)
        raise exceptions.InvalidFormatError

    shared_header = jwe_dict.get('unprotected', {})

    if 'protected' in jwe_dict:
        protected_b64 = jwe_dict['protected']
        protected_dict = json.loads(utils.to_string(utils.base64url_decode(protected_b64)))
        shared_header.update(protected_dict)

    return shared_header


def normalize_claims(raw_claims, issuer=None):
    """
    Return a set of claims based on the provided claim set that includes reasonable defaults
    for required claims.

    Note that the claims may be in the form of a valid JWE, in which case the inner values
    may be inspected.

    :param raw_claims: Initial set of claims, may or may not include required claims
    :type raw_claims: dict
    :param issuer: (optional) identifier of the identity creating the message
    :type str:
    :returns: filled-out claims
    :rtype: dict
    """
    exp = None
    nbf = None
    nonce = None

    if is_jwe(raw_claims):
        headers = get_jwe_shared_header(raw_claims)
        exp = headers.get('exp', exp)
        nbf = headers.get('nbf', nbf)
        nonce = headers.get('jti', nonce)
        if not issuer:
            issuer = headers.get('iss')
    else:
        exp = raw_claims.get('exp', exp)
        nbf = raw_claims.get('nbf', nbf)
        nonce = raw_claims.get('jti', nonce)
        if not issuer:
            issuer = raw_claims.get('iss')

    if exp and not nonce:
        # use message expiration for nonce expiration
        exp_dt = datetime.fromtimestamp(exp, tz.tzutc())
        nonce = nonces.make_nonce(exp_dt)
    elif nonce and (nonce[:3] == '002') and not exp:
        # use >v1 nonce expiration for message expiration
        try:
            exp = utils.to_timestamp(nonce[3:-6])
        except:
            logger.warning('unable to parse jti for nonce exp, using default, jti=%s', nonce)

    now = int(time.time())
    default_exp_ts = (now + TOKEN_EXPIRATION_TIME_SEC)
    default_exp_dt = datetime.fromtimestamp(default_exp_ts, tz.tzutc())

    claims = {
        'jti': nonce or nonces.make_nonce(default_exp_dt),
        'nbf': nbf or now,
        'exp': exp or default_exp_ts,
    }
    if issuer:
        claims['iss'] = issuer

    claims.update(raw_claims)

    return claims


def as_dict(msg, json_decoder=json.loads):
    """
    Unpack a message (if necessary) into its dictionary form.

    :param msg: message to convert
    :type msg: str or dict
    :param json_decoder: a function to decode JSON into a :py:class:`dict`. Defaults to `json.loads`
    :returns: the message in dictionary form
    :rtype: dict
    """

    if not isinstance(msg, dict):
        msg = json_decoder(utils.to_string(msg))

    return msg



from . import keychain
from . import service
from . import session
from . import jwts
from . import jwes
from . import nonces
from . import utils


__all__ = (keychain, service, session, jwts, jwes, nonces, utils)



# -*- coding: utf-8 -*-

"""
Provides useful functions for interacting with the Neustar TDI Core API, including creation of
keys, etc.
"""
from __future__ import unicode_literals

import os
import re
import logging

from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.serialization \
    import Encoding, PrivateFormat, NoEncryption

from .keychain import Keypair
from . import jwts
from . import file_adapter
from . import symcrypt

logger = logging.getLogger(__name__)


AUTHENTICATION_ENDPOINT = 'http://developer-portal.oneid.com/api/{project}/authenticate'

_BACKEND = default_backend()


class ServiceCreator(object):
    """
    Read yaml file and add methods dynamically from file
    Created by Session
    """
    def create_service_class(self, service_name, service_model, session, **kwargs):
        """
        Service Model is either user, server or edge_device
        """
        class_attrs = self._create_methods(service_model, **kwargs)
        cls = type(str(service_name), (BaseService,), class_attrs)

        return cls(session, kwargs.get('project_credentials'))

    def _create_methods(self, service_model, **kwargs):
        """
        :param service_model:
        :return: Dictionary of class attributes
        """
        base_url = os.environ.get('ONEID_API_SERVER_BASE_URL', kwargs.get('base_url', ''))

        methods = dict()
        for method_name, method_values in service_model.items():
            required_jwt = []
            all_jwt = []
            required_url = []
            all_url = []

            for arg_name, arg_properties in method_values['arguments'].items():
                if arg_properties['location'] == 'jwt':
                    all_jwt.append(arg_name)
                    if arg_properties['required'] is True:
                        required_jwt.append(arg_name)
                if arg_properties['location'] == 'url':
                    all_url.append(arg_name)
                    if arg_properties['required'] is True:
                        required_url.append(arg_name)

            absolute_url = '{base}{endpoint}'.format(base=base_url,
                                                     endpoint=method_values['endpoint'])

            methods[method_name] = self._create_api_method(method_name,
                                                           absolute_url,
                                                           method_values['method'],
                                                           all_body_args=all_jwt,
                                                           required_body_args=required_jwt,
                                                           all_url_args=all_url,
                                                           required_url_args=required_url,
                                                           )
        return methods

    def _create_api_method(self, name,
                           endpoint, http_method,
                           all_body_args, required_body_args,
                           all_url_args, required_url_args,
                           ):
        """
        Add methods to session dynamically from yaml file

        :param method_name: method that will be called
        """
        def _api_call(self, *args, **kwargs):
            if kwargs.get('body') is None:
                # if the body isn't specified, check for
                # required body arguments
                for required in required_body_args:
                    if required not in kwargs:
                        raise TypeError('Missing Required Keyword Argument:'
                                        ' %s' % required)
                kwargs.update(body_args=all_body_args)
            for required in required_url_args:
                if required not in kwargs:
                    raise TypeError('Missing Required URL Argument: %s' % required)
            return self._make_api_request(endpoint, http_method, **kwargs)

        _api_call.__name__ = str(name)
        return _api_call


class BaseService(object):
    """
    Dynamically loaded by data files.
    """
    def __init__(self, session, project_credentials=None):
        """
        Create a new Service

        :param session: :class:`oneid.session.Session` instance
        """
        self.session = session

        self.project_credentials = None
        if hasattr(self.session, 'project_credentials') and self.session.project_credentials:
            self.project_credentials = self.session.project_credentials

        self.identity = self.session.identity_credentials.id
        self.credentials = self.session.identity_credentials

        if self.project_credentials and self.project_credentials.id:
            self.project_id = self.project_credentials.id

    def _format_url(self, url_template, **kwargs):
        """
        Url from yaml may require formatting

        :Example:

            /project/{project_id}
            >>> /project/abc-123

        :param url_template: url with arguments that need replaced by vars
        :param params: Dictionary lookup to replace url arguments with
        :return: absolute url
        """
        encoded_params = dict()
        url_args = re.findall(r'{(\w+)}', url_template)
        for url_arg in url_args:
            if url_arg in kwargs:
                encoded_params[url_arg] = kwargs[url_arg]
            elif hasattr(self, url_arg):
                # Check if the argument is a class attribute (i.e. project_id)
                encoded_params[url_arg] = getattr(self, url_arg)
            else:
                raise TypeError('Missing URL argument %s' % url_arg)
        return url_template.format(**encoded_params)

    def _make_api_request(self, endpoint, http_method, **kwargs):
        """
        Convenience method to make HTTP requests and handle responses/error codes

        :param endpoint: URL to the resource
        :param http_method: HTTP method, GET, POST, PUT, DELETE
        :param kwargs: Params to pass to the body or url
        """
        # Split the params based on their type (url or jwt)
        url = self._format_url(endpoint, **kwargs)

        if kwargs.get('body_args'):
            claims = {arg: kwargs[arg] for arg in kwargs.get('body_args')}
            jwt = jwts.make_jwt(claims, self.credentials.keypair)
            return self.session.service_request(http_method, url, body=jwt)
        else:
            # Replace the entire body with kwargs['body'] (if present)
            return self.session.service_request(http_method, url, body=kwargs.get('body'))


def create_secret_key(output=None):
    """
    Create a secret key and save it to a secure location

    :param output: Path to save the secret key
    :return: oneid.keychain.Keypair
    """
    secret_key = ec.generate_private_key(ec.SECP256R1(), _BACKEND)
    secret_key_bytes = secret_key.private_bytes(Encoding.PEM, PrivateFormat.PKCS8, NoEncryption())

    # Save the secret key bytes to a secure file
    if output and file_adapter.file_directory_exists(output):
        file_adapter.write_file(output, secret_key_bytes)

    return Keypair.from_secret_pem(key_bytes=secret_key_bytes)


def create_aes_key():
    """
    Create an AES256 key for symmetric encryption

    :return: Encryption key bytes
    """
    return symcrypt.create_aes_key()


def encrypt_attr_value(attr_value, aes_key, legacy_support=True):
    """
    Convenience method to encrypt attribute properties

    :param attr_value: plain text (string or bytes) that you want encrypted
    :param aes_key: symmetric key to encrypt attribute value with
    :return: Dictionary (Flattened JWE) with base64-encoded ciphertext and base64-encoded iv
    """
    return symcrypt.aes_encrypt(attr_value, aes_key, legacy_support)


def decrypt_attr_value(attr_ct, aes_key):
    """
    Convenience method to decrypt attribute properties

    :param attr_ct: Dictionary (may be a Flattened JWE) with base64-encoded
        ciphertext and base64-encoded iv
    :param aes_key: symmetric key to decrypt attribute value with
    :return: plaintext bytes
    """
    return symcrypt.aes_decrypt(attr_ct, aes_key)


def key_wrap(wrapping_key, key_to_wrap):
    return symcrypt.key_wrap(wrapping_key, key_to_wrap)


def key_unwrap(wrapping_key, wrapped_key):
    return symcrypt.key_unwrap(wrapping_key, wrapped_key)




"""
A Keypair is used to sign and verify signatures

Keys should be kept in a secure storage enclave.
"""
from __future__ import division

import struct
import logging

from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.serialization import load_pem_private_key, \
    load_pem_public_key, load_der_private_key, load_der_public_key

from cryptography.hazmat.primitives.asymmetric.utils \
    import decode_dss_signature, encode_dss_signature
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.serialization \
    import Encoding, PublicFormat, PrivateFormat, NoEncryption
from cryptography.hazmat.primitives.kdf.concatkdf import ConcatKDFHash
from cryptography.utils import int_to_bytes, int_from_bytes

from . import symcrypt, utils, exceptions, file_adapter

KEYSIZE = 256
KEYSIZE_BYTES = (KEYSIZE // 8)

logger = logging.getLogger(__name__)

_BACKEND = default_backend()


class Credentials(object):
    """
    Container for User/Server/Device Encryption Key, Signing Key, Identity


    :ivar identity: UUID of the identity.
    :ivar keypair: :class:`~oneid.keychain.BaseKeypair` instance.
    """
    def __init__(self, identity, keypair):
        """

        :param identity: uuid of the entity
        :param keypair: :py:class:`~oneid.keychain.BaseKeypair` instance
        """
        self.id = identity

        if not isinstance(keypair, BaseKeypair):
            raise ValueError('keypair must be a oneid.keychain.BaseKeypair instance')

        self.keypair = keypair


class ProjectCredentials(Credentials):
    def __init__(self, project_id, keypair, encryption_key):
        """
        Adds an encryption key

        :param project_id: oneID project UUID
        :param keypair: :py:class:`~oneid.keychain.BaseKeypair`
        :param encryption_key: AES key used to encrypt messages
        """
        super(ProjectCredentials, self).__init__(project_id, keypair)
        self._encryption_key = encryption_key

    def encrypt(self, plain_text):
        """
        Encrypt plain text with the project encryption key.

        :param plain_text: String or bytes to encrypt with project encryption key.
        :returns: Dictionary with cipher text and encryption params.
        """
        return symcrypt.aes_encrypt(plain_text, self._encryption_key)

    def decrypt(self, cipher_text):
        """
        Decrypt cipher text that was encrypted with the project encryption key

        :param cipher_text: Encrypted dict as returned by :py:encrypt:
        :returns: plain text
        :return_type: bytes
        """
        return symcrypt.aes_decrypt(cipher_text, self._encryption_key)


class BaseKeypair(object):
    """
    Generic :py:class:`~oneid.keychain.Keypair` functionality.

    Callers can subclass this to mimic or proxy
    :py:class:`~oneid.keychain.Keypair`\s
    """
    def __init__(self, *args, **kwargs):
        self.identity = kwargs.get('identity')

    @property
    def public_key_der(self):
        raise NotImplementedError

    @property
    def public_key_pem(self):
        raise NotImplementedError

    @property
    def secret_as_der(self):
        raise NotImplementedError

    @property
    def secret_as_pem(self):
        raise NotImplementedError

    @property
    def jwk(self):
        raise NotImplementedError

    @property
    def jwk_private(self):
        raise NotImplementedError

    @property
    def jwk_public(self):
        raise NotImplementedError

    def verify(self, payload, signature):
        raise NotImplementedError

    def sign(self, payload):
        raise NotImplementedError

    def ecdh(self, peer_keypair, algorithm='A256GCM', party_u_info=None, party_v_info=None):
        raise NotImplementedError

    def save(self, *args, **kwargs):
        """
        Save a key.
        Should be overridden and saved to secure storage

        :param args:
        :param kwargs:
        :return: Bool Success
        """
        raise NotImplementedError

    def _calc_otherinfo(self, algorithm, party_u_info, party_v_info):
        """
        Broken out so testing can override to inject CAVP vector data
        """
        return (
            _len_bytes(algorithm) +
            _len_bytes(party_u_info) +
            _len_bytes(party_v_info) +
            utils.to_bytes(struct.pack(">I", 256))
        )

    def _derive_ecdh(self, raw_key, otherinfo):
        ckdf = ConcatKDFHash(
            algorithm=hashes.SHA256(),
            length=32,
            otherinfo=otherinfo,
            backend=_BACKEND,
        )
        return ckdf.derive(raw_key)


class Keypair(BaseKeypair):
    def __init__(self, *args, **kwargs):
        """
        :param kwargs: Pass secret key bytes
        """
        super(Keypair, self).__init__(*args, **kwargs)

        self._private_key = None
        self._public_key = None

        if kwargs.get('secret_bytes') and \
                isinstance(kwargs['secret_bytes'], ec.EllipticCurvePrivateKey):
            self._load_secret_bytes(kwargs['secret_bytes'])

    def _load_secret_bytes(self, secret_bytes):
        self._private_key = secret_bytes

    @property
    def secret_as_der(self):
        """
        Write out the private key as a DER format

        :return: DER encoded private key
        """
        secret_der = self._private_key.private_bytes(
            Encoding.DER, PrivateFormat.PKCS8, NoEncryption()
        )

        return secret_der

    @property
    def secret_as_pem(self):
        """
        Write out the private key as a PEM format

        :return: Pem Encoded private key
        """
        return self._private_key.private_bytes(Encoding.PEM, PrivateFormat.PKCS8, NoEncryption())

    @classmethod
    def from_secret_pem(cls, key_bytes=None, path=None):
        """
        Create a :class:`~oneid.keychain.Keypair` from a PEM-formatted private ECDSA key

        :return: :class:`~oneid.keychain.Keypair` instance
        """
        if key_bytes:
            secret_bytes = load_pem_private_key(utils.to_bytes(key_bytes), None, _BACKEND)
            return cls(secret_bytes=secret_bytes)

        if file_adapter.file_exists(path):
            with file_adapter.read_file(path) as pem_data:
                secret_bytes = load_pem_private_key(pem_data, None, _BACKEND)
                return cls(secret_bytes=secret_bytes)

    @classmethod
    def from_public_pem(cls, key_bytes=None, path=None):
        """
        Create a :class:`~oneid.keychain.Keypair` from a PEM-formatted public ECDSA key

        Note that this keypair will not be capable of signing, only verifying.

        :return: :class:`~oneid.keychain.Keypair` instance
        """
        ret = None
        public_bytes = None

        if key_bytes:
            public_bytes = utils.to_bytes(key_bytes)
        elif file_adapter.file_exists(path):
            with file_adapter.read_file(path) as pem_data:
                public_bytes = pem_data

        if public_bytes:
            ret = cls()
            ret._public_key = load_pem_public_key(public_bytes, _BACKEND)

        return ret

    @classmethod
    def from_secret_der(cls, der_key):
        """
        Read a der_key, convert it a private key

        :param path: der formatted key
        :return:
        """
        secret_bytes = load_der_private_key(der_key, None, _BACKEND)
        return cls(secret_bytes=secret_bytes)

    @classmethod
    def from_public_der(cls, public_key):
        """
        Given a DER-format public key, convert it into a token to
        validate signatures

        :param public_key: der formatted key
        :return: :class:`~oneid.keychain.Keypair` instance
        """
        pub = load_der_public_key(public_key, _BACKEND)

        new_token = cls()
        new_token._public_key = pub

        return new_token

    @classmethod
    def from_jwk(cls, jwk):
        """
        Create a :py:class:`~oneid.keychain.Keypair` from a JWK

        :param jwk: oneID-standard JWK
        :return: :py:class:`~oneid.keychain.Keypair` instance
        :raises InvalidFormatError: if not a valid JWK
        """
        if jwk['kty'] != 'EC' or jwk['crv'] != 'P-256':
            raise exceptions.InvalidFormatError

        public_numbers = ec.EllipticCurvePublicNumbers(
            x=int_from_bytes(utils.base64url_decode(jwk['x']), 'big'),
            y=int_from_bytes(utils.base64url_decode(jwk['y']), 'big'),
            curve=ec.SECP256R1(),
        )

        ret = cls()
        ret._public_key = public_numbers.public_key(_BACKEND)

        if 'd' in jwk:
            private_numbers = ec.EllipticCurvePrivateNumbers(
                private_value=int_from_bytes(utils.base64url_decode(jwk['d']), 'big'),
                public_numbers=public_numbers,
            )
            ret._private_key = private_numbers.private_key(_BACKEND)

        if 'kid' in jwk:
            ret.identity = jwk['kid']

        return ret

    @property
    def jwk(self):
        """
        The keys as a JSON Web Key (JWK)
        Private key will be included only if present

        :return: oneID-standard JWK
        """
        return self.get_jwk(True)

    @property
    def jwk_public(self):
        """
        The public key as a JSON Web Key (JWK)

        :return: oneID-standard JWK
        """
        return self.get_jwk(False)

    @property
    def jwk_private(self):
        """
        The private key as a JSON Web Key (JWK)

        :return: oneID-standard JWK
        :raises InvalidFormatError: if not a private key
        """
        if not self._private_key:
            raise exceptions.InvalidFormatError
        return self.get_jwk(True)

    def get_jwk(self, include_secret):
        public_numbers = self.public_key.public_numbers()
        ret = {
          "kty": "EC",
          "crv": "P-256",
          "x": utils.to_string(utils.base64url_encode(int_to_bytes(public_numbers.x))),
          "y": utils.to_string(utils.base64url_encode(int_to_bytes(public_numbers.y))),
        }

        if self.identity:
            ret['kid'] = str(self.identity)

        if self._private_key and include_secret:
            private_numbers = self._private_key.private_numbers()
            d = int_to_bytes(private_numbers.private_value)
            ret['d'] = utils.to_string(utils.base64url_encode(d))

        return ret

    def verify(self, payload, signature):
        """
        Verify that the token signed the data

        :type payload: String
        :param payload: message that was signed and needs verified
        :type signature: Base64 URL Safe
        :param signature: Signature that can verify the sender\'s identity and payload
        :return:
        """
        raw_sig = utils.base64url_decode(signature)
        sig_r_bin = raw_sig[:len(raw_sig)//2]
        sig_s_bin = raw_sig[len(raw_sig)//2:]

        sig_r = int_from_bytes(sig_r_bin, 'big')
        sig_s = int_from_bytes(sig_s_bin, 'big')

        sig = encode_dss_signature(sig_r, sig_s)
        self.public_key.verify(sig, utils.to_bytes(payload), ec.ECDSA(hashes.SHA256()))
        return True

    def sign(self, payload):
        """
        Sign a payload

        :param payload: String (usually jwt payload)
        :return: URL safe base64 signature
        """
        signature = self._private_key.sign(utils.to_bytes(payload), ec.ECDSA(hashes.SHA256()))

        r, s = decode_dss_signature(signature)

        br = int_to_bytes(r, KEYSIZE_BYTES)
        bs = int_to_bytes(s, KEYSIZE_BYTES)
        str_sig = br + bs
        b64_signature = utils.base64url_encode(str_sig)
        return b64_signature

    def ecdh(self, peer_keypair, algorithm='A256GCM', party_u_info=None, party_v_info=None):
        """
        Derive a shared symmetric key for encrypting data to a given recipient

        :param peer_keypair: Public key of the recipient
        :type peer_keypair: :py:class:`~oneid.keychain.Keypair`
        :param algorithm: The algorithm associated with the operation (defaults to 'A256GCM')
        :type algorithm: str
        :param party_u_info: shared identifying information about the sender (optional)
        :type party_u_info: str or bytes
        :param party_v_info: shared identifying information about the recipient (optional)
        :type party_v_info: str or bytes
        :returns: a 256-bit encryption key
        :return_type: bytes
        :raises InvalidFormatError: if self is not a private key
        """
        if not self._private_key:
            raise exceptions.InvalidFormatError
        raw_key = self._raw_ecdh(peer_keypair)
        otherinfo = self._calc_otherinfo(algorithm, party_u_info, party_v_info)
        ret = self._derive_ecdh(raw_key, otherinfo)

        del raw_key

        return ret

    @property
    def public_key(self):
        """
        If the private key is defined, generate the public key

        :return:
        """
        if self._public_key:
            return self._public_key
        elif self._private_key:
            return self._private_key.public_key()

    @property
    def public_key_der(self):
        """
        DER formatted public key

        :return: Public Key in DER format
        """
        return self.public_key.public_bytes(Encoding.DER, PublicFormat.SubjectPublicKeyInfo)

    @property
    def public_key_pem(self):
        """
        PEM formatted public key

        :return: Public Key in PEM format
        """
        return self.public_key.public_bytes(Encoding.PEM, PublicFormat.SubjectPublicKeyInfo)

    def _raw_ecdh(self, peer_keypair):
        return self._private_key.exchange(ec.ECDH(), peer_keypair.public_key)


def _len_bytes(data):
    if not data:
        return utils.to_bytes('')
    return utils.to_bytes(struct.pack(">I", len(data))) + utils.to_bytes(data)




import os
import contextlib


def join_paths(*paths):
    return os.path.join(*paths)


def file_exists(path):
    return os.path.exists(path)


def file_directory_exists(filename):
    return os.path.exists(os.path.dirname(filename))


def prepare_directory(dirname):
    if not os.path.exists(dirname):
        os.makedirs(dirname)


def prepare_file_directory(filename):
    prepare_directory(os.path.dirname(filename))


@contextlib.contextmanager
def read_file(filename, binary):
    mode = 'r' + ('b' if binary else '')

    with open(filename, mode) as f:
        yield f.read()


def write_file(filename, data, binary):
    mode = 'w' + ('b' if binary else '')

    with open(filename, mode) as f:
        f.write(data)



# -*- coding: utf-8 -*-

"""
Provides useful functions for dealing with JWTs and JWSs

Based on the
`JWT <https://tools.ietf.org/html/rfc7519/>`_ and
`JWS <https://tools.ietf.org/html/rfc7515/>`_
IETF RFCs.

"""
from __future__ import unicode_literals

import six
import collections
import json
import time
import logging

from datetime import datetime
from dateutil import tz

from . import jose, nonces, utils, exceptions

logger = logging.getLogger(__name__)


RESERVED_HEADERS = [
    'typ',
    'alg',
    'kid',
    'sidx'
]
MINIMAL_JWT_HEADER = {
    'typ': 'JWT',
    'alg': 'ES256',
}
MINIMAL_JSON_JWS_HEADER = {
    'typ': 'JOSE+JSON',
    'alg': 'ES256',
}
TOKEN_EXPIRATION_TIME_SEC = nonces.DEFAULT_NONCE_EXPIRY_SECONDS
TOKEN_NOT_BEFORE_LEEWAY_SEC = (2 * 60)   # two minutes
TOKEN_EXPIRATION_LEEWAY_SEC = (3)      # three seconds


def make_jwt(raw_claims, keypair, json_encoder=json.dumps):
    """
    Convert claims into JWT

    :param raw_claims: payload data that will be converted to json
    :type raw_claims: dict
    :param keypair: :py:class:`~oneid.keychain.Keypair` to sign the request
    :param json_encoder: a function to encode a :py:class:`dict` into JSON. Defaults to `json.dumps`
    :return: JWT
    """
    if not isinstance(raw_claims, dict):
        raise TypeError('dict required for claims, type=' +
                        str(type(raw_claims)))

    claims = jose.normalize_claims(raw_claims, keypair.identity)
    claims_serialized = json_encoder(claims)
    claims_b64 = utils.to_string(utils.base64url_encode(claims_serialized))

    header = {}
    header.update(MINIMAL_JWT_HEADER)
    if keypair.identity:
        header['kid'] = keypair.identity
    header_b64 = utils.to_string(utils.base64url_encode(json_encoder(header)))

    payload = '{header}.{claims}'.format(header=header_b64, claims=claims_b64)

    signature = utils.to_string(keypair.sign(payload))

    return '{payload}.{sig}'.format(payload=payload, sig=signature)


def verify_jwt(jwt, keypair=None, json_decoder=json.loads):
    """
    Convert a JWT back to it's claims, if validated by the :py:class:`~oneid.keychain.Keypair`

    :param jwt: JWT to verify and convert
    :type jwt: str or bytes
    :param keypair: :py:class:`~oneid.keychain.Keypair` to verify the JWT
    :type keypair: :py:class:`~oneid.keychain.Keypair`
    :param json_decoder: a function to decode JSON into a :py:class:`dict`. Defaults to `json.loads`
    :returns: claims
    :rtype: dict
    :raises: :py:class:`~oneid.exceptions.InvalidFormatError` if not a valid JWT
    :raises: :py:class:`~oneid.exceptions.InvalidAlgorithmError` if unsupported algorithm specified
    :raises: :py:class:`~oneid.exceptions.InvalidClaimsError` if missing or invalid claims,
        including expiration, re-used nonce, etc.
    :raises: :py:class:`~oneid.exceptions.InvalidSignatureError` if signature is not valid
    """
    jwt = utils.to_string(jwt)

    if not jose.is_compact_jws(jwt):
        logger.debug('Given JWT doesnt match pattern: %s', jwt)
        raise exceptions.InvalidFormatError

    try:
        header_json, claims_json, signature = [
            utils.base64url_decode(p) for p in jwt.split('.')]
    except:
        logger.debug('invalid JWT, error splitting/decoding: %s',
                     jwt, exc_info=True)
        raise exceptions.InvalidFormatError

    header = _verify_jose_header(
        utils.to_string(header_json), True, json_decoder)
    claims = _verify_claims(utils.to_string(claims_json), json_decoder)

    if keypair:
        try:
            keypair.verify(*(jwt.rsplit('.', 1)))
        except:
            logger.debug('invalid signature, header=%s, claims=%s',
                         header, claims, exc_info=True)
            raise exceptions.InvalidSignatureError

    if 'jti' in claims:
        nonces.burn_nonce(claims['jti'])

    return claims


def make_jws(raw_claims, ordered_keypairs, multiple_sig_headers=None, json_encoder=json.dumps):
    """
    Convert claims into JWS format (compact or JSON)

    :param raw_claims: payload data that will be converted to json
    :type raw_claims: dict
    :param ordered_keypairs:
        :py:class:`~oneid.keychain.Keypair`\s to sign the request with (in signing order,
        with ordered_keypairs[0] the first the sign the JWS)
    :type ordered_keypairs: list
    :param json_encoder: a function to encode a :py:class:`dict` into JSON. Defaults to `json.dumps`
    :param multiple_sig_headers:
        optional list of headers for associated keypairs with the same list index
    :type multiple_sig_headers: list
    :return: JWS
    """

    if not isinstance(ordered_keypairs, collections.Iterable):
        ordered_keypairs = [ordered_keypairs]

    multiple_sig_headers = _validated_headers(ordered_keypairs, multiple_sig_headers)
    multiple_sig_headers = _add_signature_indexes(multiple_sig_headers)

    claims = jose.normalize_claims(raw_claims)
    claims_serialized = json_encoder(claims)
    claims_b64 = utils.to_string(utils.base64url_encode(claims_serialized))

    ret = {
        'payload': claims_b64,
        'signatures': [_protected_signature(claims_b64, k, json_encoder=json_encoder, header=h)
                       for (k, h) in zip(ordered_keypairs, multiple_sig_headers)]
    }

    return json_encoder(ret)


def extend_jws_signatures(
    jws, ordered_keypairs, default_jwt_kid=None, multiple_sig_headers=None,
    json_encoder=json.dumps, json_decoder=json.loads
):
    """
    Add signatures to an existing JWS (or JWT)

    :param jws: existing JWS (Compact or JSON) or JWT
    :type jws: str
    :param ordered_keypairs:
        :py:class:`~oneid.keychain.Keypair`\s to sign the request with (in signing order,
        with ordered_keypairs[0] the first the sign the JWS)
    :type ordered_keypairs: list
    :param default_jwt_kid: (optional) value for 'kid' header field if passing a JWT without one
    :type default_jwt_kid: str
    :param json_encoder: a function to encode a :py:class:`dict` into JSON. Defaults to `json.dumps`
    :param json_decoder: a function to decode JSON into a :py:class:`dict`. Defaults to `json.loads`
    :param multiple_sig_headers:
        optional list of headers for associated keypairs with the same list index
    :type multiple_sig_headers: list
    :return: JWS
    """
    ret = _jws_as_dict(jws, default_jwt_kid, json_decoder)
    payload = ret['payload']

    if not isinstance(ordered_keypairs, collections.Iterable):
        ordered_keypairs = [ordered_keypairs]

    multiple_sig_headers = _validated_headers(ordered_keypairs, multiple_sig_headers)

    existing_headers = [_get_signature_header(s, json_decoder) for s in ret.get('signatures', [])]

    signature_indexes = [h.get('sidx', None) for h in existing_headers]
    kids = [h.get('kid', None) for h in existing_headers]

    if signature_indexes.count(None) <= 1:
        signature_indexes = [s for s in signature_indexes if s is not None]
        if len(signature_indexes) == 0:
            signature_offset = 0
        else:
            signature_offset = max(signature_indexes)
        multiple_sig_headers = _add_signature_indexes(multiple_sig_headers,
                                                      offset=(signature_offset + 1))

    for sig_header in multiple_sig_headers:
        sig_header['kids'] = kids
        sig_header['sidxs'] = signature_indexes

    new_signatures = [_protected_signature(payload, k, json_encoder=json_encoder, header=h)
                      for (k, h) in zip(ordered_keypairs, multiple_sig_headers)]

    ret['signatures'].extend(new_signatures)

    return json_encoder(ret)


def remove_jws_signatures(jws, kids_to_remove, json_encoder=json.dumps, json_decoder=json.loads):
    """
    Remove signatures from an existing JWS

    :param jws: existing JWS (JSON format only)
    :type jws: str
    :param kids_to_remove: Keypair identities to remove
    :type kids_to_remove: list
    :param json_encoder: a function to encode a :py:class:`dict` into JSON. Defaults to `json.dumps`
    :param json_decoder: a function to decode JSON into a :py:class:`dict`. Defaults to `json.loads`
    :return: JWS (may have empty signature list if last one removed)
    """
    jws_dict = json_decoder(utils.to_string(jws))

    if isinstance(kids_to_remove, six.string_types + (six.binary_type,)):
        kids_to_remove = [kids_to_remove]

    return json_encoder({
        'payload': jws_dict['payload'],
        'signatures': [
            sig for sig in jws_dict['signatures']
            if _get_kid_for_signature(sig, None, json_decoder) not in kids_to_remove
        ]
    })


def get_jws_key_ids(jws, default_kid=None, json_decoder=json.loads, ordered=False):
    """
    Extract the IDs of the keys used to sign a given JWS

    :param jws: JWS to get key IDs from
    :type jws: str or bytes
    :param default_kid: Value to use for looking up keypair if no `kid` found
                    in a given signature header, as may happen when extending a JWT
    :type default_kid: str
    :param json_decoder: a function to decode JSON into a :py:class:`dict`. Defaults to `json.loads`
    :returns: key IDs
    :rtype: list
    :param ordered:
        Bool if the key IDs should be returned in the order they signed the JWS.
        If this cannot be resolved, an exception is thrown.
    :raises: :py:class:`~oneid.exceptions.InvalidFormatError`: if not a valid JWS
    """

    jws = utils.to_string(jws)

    if jose.is_compact_jws(jws):
        header_b64, claims_b64, _ = jws.split('.')
        header = json_decoder(utils.to_string(
            utils.base64url_decode(header_b64)))
        claims = json_decoder(utils.to_string(
            utils.base64url_decode(claims_b64)))
        kid = header.get('kid', claims.get('iss', default_kid))
        return [{'kid': kid, 'kids': [], 'sidxs': []}] if kid else []

    try:
        jws = json_decoder(jws)
        payload_json = utils.to_string(
            utils.base64url_decode(jws.get('payload', '')))
        payload = json_decoder(payload_json)
        default_kid = payload and payload.get(
            'iss', default_kid) or default_kid
    except:
        logger.debug('error parsing JWS', exc_info=True)
        raise exceptions.InvalidFormatError

    headers = [_get_signature_header(s, json_decoder) for s in jws['signatures']]

    if ordered:
        signature_indexes = [h.get('sidx', None) for h in headers]
        if signature_indexes.count(None) > 1:
            raise exceptions.InvalidSignatureIndexes("Some signature headers are missing indexes")
        if len(signature_indexes) != len(set(signature_indexes)):
            raise exceptions.InvalidSignatureIndexes("Duplicate signature indexes")

        headers = sorted(headers, key=lambda h: h.get('sidx', 0))

    return [
        {
            'kid': h.get('kid', default_kid),
            'kids': h.get('kids', []),
            'sidxs': h.get('sidxs', []),
        }
        for h in headers
    ]


def verify_jws(jws, keypairs=None, verify_all=True, default_kid=None, json_decoder=json.loads):
    """
    Convert a JWS back to it's claims, if validated by a set of
    required :py:class:`~oneid.keychain.Keypair`\s

    :param jws: JWS to verify and convert
    :type jws: str or bytes
    :param keypairs: :py:class:`~oneid.keychain.Keypair`\s to verify the JWS with.
                    Must include one for each specified in the JWS headers' `kid` values.
    :type keypairs: list
    :param verify_all: If True (default), all keypairs must validate a signature.
                    If False, only one needs to.
                    If any fail to validate, the JWS is still not validated.
                    This allows the caller to send multiple keys that _might_ have
                    corresponding signatures, without requiring that _all_ do.
    :type verify_all: bool
    :param default_kid: Value to use for looking up keypair if no `kid` found
                    in a given signature header, as may happen when extending a JWT
    :type default_kid: str
    :param json_decoder: a function to decode JSON into a :py:class:`dict`. Defaults to `json.loads`
    :returns: claims
    :rtype: dict
    :raises: :py:class:`~oneid.exceptions.InvalidFormatError`: if not a valid JWS
    :raises: :py:class:`~oneid.exceptions.InvalidAlgorithmError`: if unsupported algorithm specified
    :raises: :py:class:`~oneid.exceptions.InvalidClaimsError`: if missing or invalid claims,
        including expiration, re-used nonce, etc.
    :raises: :py:class:`~oneid.exceptions.InvalidSignatureError`: if any relevant signature
        is not valid
    """

    if keypairs and not isinstance(keypairs, collections.Iterable):
        keypairs = [keypairs]

    jws = utils.to_string(jws)

    if jose.is_compact_jws(jws):

        if verify_all and keypairs and len(keypairs) != 1:
            raise exceptions.InvalidSignatureError(
                'Compact JWS found but multiple signatures required'
            )

        return verify_jwt(jws, keypairs and keypairs[0])

    jws = json_decoder(jws)

    if 'payload' not in jws or 'signatures' not in jws:
        raise exceptions.InvalidFormatError

    claims = _verify_claims(utils.to_string(
        utils.base64url_decode(jws['payload'])), json_decoder)

    if keypairs:
        _verify_jws_signatures(jws, keypairs, verify_all,
                               default_kid, json_decoder)

    if 'jti' in claims:
        nonces.burn_nonce(claims['jti'])

    return claims


def get_jws_headers(jws, json_decoder=json.loads):
    """
    Extract the headers of the signatures used to sign a given JWS

    :param jws: JWS to get headers from
    :type jws: str or bytes
    :param json_decoder: a function to decode JSON into a :py:class:`dict`. Defaults to `json.loads`
    :returns: headers
    :rtype: list
    :raises: :py:class:`~oneid.exceptions.InvalidFormatError`: if not a valid JWS
    """
    jws = utils.to_string(jws)
    try:
        jws = json_decoder(jws)
    except:
        raise exceptions.InvalidFormatError

    return [_get_signature_header(s, json_decoder) for s in jws.get('signatures', [])]


# def _extract_claim(jws, claim):
#
#     if jose.is_compact_jws(jws):
#         claims_b64 = jws.split('.')[1]
#     else:
#         jws_dict = json.loads(jws)
#         claims_b64 = jws_dict.get('payload', '')
#
#     claims = utils.base64url_decode(claims_b64) or {}
#
#     return claims.get(claim)


def _protected_signature(claims_b64, keypair, json_encoder=json.dumps,
                         header=None):
    if not keypair.identity:
        logger.debug('Missing Keypair.identity')
        raise exceptions.InvalidKeyError

    header = header or {}
    header['kid'] = keypair.identity
    header.update(MINIMAL_JSON_JWS_HEADER)
    header_b64 = utils.to_string(utils.base64url_encode(json_encoder(header)))
    to_sign = '{header}.{claims}'.format(header=header_b64, claims=claims_b64)

    sig = utils.to_string(keypair.sign(to_sign))

    return {
        'protected': header_b64,
        'signature': sig
    }


def _validated_headers(keypairs, multiple_sig_headers):

    number_keys = len(keypairs)

    if multiple_sig_headers is not None and number_keys != len(multiple_sig_headers):
        raise exceptions.KeyHeaderMismatch(
            'The number of Keypairs and Headers must be equal.')

    multiple_sig_headers = multiple_sig_headers or [{} for _ in range(number_keys)]

    for header in multiple_sig_headers:
        invalid_header = next((k for k in header.keys() if k in RESERVED_HEADERS), False)
        if invalid_header:
            raise exceptions.ReservedHeader(invalid_header + " is an invalid JWS header")

    return multiple_sig_headers


def _add_signature_indexes(multiple_sig_headers, offset=0):
    def add_to_dictionary(d, k, v):
        d[k] = v
        return d

    return [add_to_dictionary(headers, 'sidx', idx+offset)
            for idx, headers in enumerate(multiple_sig_headers)]


def _jws_as_dict(jws, kid, json_decoder):

    jws = utils.to_string(jws)

    if not jose.is_compact_jws(jws):
        return json_decoder(jws)

    header_b64, payload, signature = jws.split('.')

    header = json_decoder(utils.to_string(utils.base64url_decode(header_b64)))
    extra_header = None

    if 'kid' not in header:
        claims = json_decoder(utils.to_string(utils.base64url_decode(payload)))
        kid = kid or claims.get('iss')
        extra_header = kid and {
            'kid': kid
        }

    ret = {
        'payload': payload,
        'signatures': [{
            'protected': header_b64,
            'signature': signature,
        }]
    }

    if extra_header:
        ret['signatures'][0]['header'] = extra_header

    return ret


def _verify_jose_header(header_json, strict_jwt, json_decoder):
    header = None
    try:
        header = json_decoder(header_json)
    except ValueError:
        logger.debug('invalid header, not valid json: %s', header_json)
        raise exceptions.InvalidFormatError
    except Exception:  # pragma: no cover
        logger.debug('unknown error verifying header: %s',
                     header, exc_info=True)
        raise

    if strict_jwt:
        keys = {k: 1 for k in header}
        for key, value in MINIMAL_JWT_HEADER.items():
            if key not in header or header.get(key, None) != value:
                logger.debug(
                    'invalid header, missing or incorrect %s: %s', key, header)
                raise exceptions.InvalidFormatError
            keys.pop(key, None)
        keys.pop('kid', None)

        if len(keys) > 0:
            logger.debug('invalid header, extra elements: %s', header)
            raise exceptions.InvalidFormatError
    else:
        if 'typ' not in header or header['typ'] not in ['JWT', 'JOSE', 'JOSE+JSON']:
            logger.debug('invalid "typ" in header: %s', header)
            raise exceptions.InvalidFormatError

        if 'alg' not in header or header['alg'] != 'ES256':
            logger.debug('invalid "alg" in header: %s', header)
            raise exceptions.InvalidAlgorithmError

    return header


def _verify_claims(payload, json_decoder):
    try:
        claims = json_decoder(payload)
    except:
        logger.debug('unknown error verifying payload: %s',
                     payload, exc_info=True)
        raise exceptions.InvalidFormatError

    now_ts = int(time.time())

    nbf = None

    if 'exp' in claims:
        exp_ts = (int(claims['exp']) + TOKEN_EXPIRATION_LEEWAY_SEC)
        if now_ts > exp_ts:
            logger.warning('Expired token, exp=%s, now_ts=%s',
                           claims['exp'], now_ts)
            raise exceptions.InvalidClaimsError

    if 'nbf' in claims:
        nbf_ts = (int(claims['nbf']) - TOKEN_NOT_BEFORE_LEEWAY_SEC)
        if now_ts < nbf_ts:
            logger.warning('Early token, nbf=%s, now_ts=%s',
                           claims['nbf'], now_ts)
            raise exceptions.InvalidClaimsError
        nbf = datetime.fromtimestamp(nbf_ts, tz.tzutc())

    if 'jti' in claims and not nonces.verify_nonce(claims['jti'], nbf):
        logger.warning('Invalid nonce: %s', claims['jti'])
        raise exceptions.InvalidClaimsError

    return claims


def _verify_jws_signatures(jws, keypairs, verify_all, default_kid, json_decoder):
    if len(jws['signatures']) == 0:
        logger.warning('No signatures found, rejecting')
        raise exceptions.InvalidSignatureError

    if verify_all and len(keypairs) != len(jws['signatures']):
        raise exceptions.KeySignatureMismatch(
            'number of keys doesn\'t match number of signatures')

    keypair_map = {str(keypair.identity): keypair for keypair in keypairs}

    if len(keypairs) != len(keypair_map):
        raise exceptions.InvalidKeyError(
            'redundant keypairs found, unable to verify')

    found_sigs = [
        _get_kid_for_signature(sig, default_kid, json_decoder)
        in keypair_map for sig in jws['signatures']
    ]

    # need at least one signature
    if not any(found_sigs):
        logger.warning('No keypairs had corresponding signatures, rejecting')
        raise exceptions.KeySignatureMismatch

    # or all
    if verify_all and not all(found_sigs):
        logger.warning('Not all keys have corresponding signatures, rejecting')
        raise exceptions.KeySignatureMismatch

    for signature in jws['signatures']:
        kid = _get_kid_for_signature(signature, default_kid, json_decoder)

        if verify_all or kid in keypair_map:
            _verify_jws_signature(
                jws['payload'], keypair_map.get(kid), signature)


def _get_kid_for_signature(signature, default_kid, json_decoder):
    header = _get_signature_header(signature, json_decoder)
    kid = header.get('kid', signature.get(
        'header', {}).get('kid', default_kid))

    if not kid:
        logger.warning(
            'invalid header in signature, missing "kid": %s', signature
        )
        raise exceptions.InvalidFormatError

    return kid


def _get_signature_header(signature, json_decoder):
    # TODO: check for overlapping keys in `protected` and `header`, merge together
    # for now, we only look for `kid` in `header`, and only if it isn't in
    # `protected`

    return _verify_jose_header(
        utils.to_string(utils.base64url_decode(signature['protected'])),
        False, json_decoder,
    )


def _verify_jws_signature(payload, keypair, signature):
    try:
        keypair.verify(
            '.'.join([signature['protected'], payload]), signature['signature'])
    except:
        logger.debug('invalid signature', exc_info=True)
        raise exceptions.InvalidSignatureError




from cryptography.exceptions import UnsupportedAlgorithm, InvalidSignature


class InvalidAuthentication(Exception):
    pass


class InvalidFormatError(ValueError):
    pass


class InvalidAlgorithmError(UnsupportedAlgorithm):

    def __init__(self, message="invalid 'alg' specified"):
        super(InvalidAlgorithmError, self).__init__(message)


class InvalidClaimsError(ValueError):
    pass


class InvalidKeyError(InvalidSignature):
    pass


class KeySignatureMismatch(InvalidSignature):
    pass


class InvalidSignatureError(InvalidSignature):
    pass


class KeyHeaderMismatch(Exception):
    pass


class ReservedHeader(Exception):
    pass


class InvalidSignatureIndexes(Exception):
    pass


class IdentityRequired(ValueError):
    pass


class InvalidRecipient(Exception):
    pass


class DecryptionFailed(Exception):
    pass




import re
import contextlib

import boto3
import botocore

import logging

from .. import utils

logger = logging.getLogger(__name__)


def join_paths(*paths):
    return '/'.join(paths)


def file_exists(path):
    return _object_exists(*_parse_path(path))


def file_directory_exists(path):
    # any key will work if the bucket exists
    return _bucket_exists(_parse_path(path)[0])


def prepare_directory(dirname):
    bucket_name, _ = _parse_path(dirname)

    if not _bucket_exists(bucket_name):
        _s3().Bucket(bucket_name).create()


def prepare_file_directory(filename):
    prepare_directory(filename)  # only need to create the bucket


@contextlib.contextmanager
def read_file(filename, binary):
    obj = _s3().Object(*_parse_path(filename))
    # data = obj.get(**settings.AWS_S3_OBJECT_PARAMS)['Body'].read()
    data = obj.get()['Body'].read()
    yield data if binary else utils.to_string(data)


def write_file(filename, data, binary):
    obj = _s3().Object(*_parse_path(filename))
    data = data if binary else utils.to_string(data)
    obj.put(Body=data, ServerSideEncryption='AES256')  # , **settings.AWS_S3_OBJECT_PARAMS)


__s3 = None


def _s3():
    global __s3

    if not __s3:
        __s3 = boto3.resource('s3')
    return __s3


def _parse_path(path):
    match = re.match(r'^s3://(?P<bucket_name>.*?)/(?P<key_name>.*)$', path)

    if not match:
        raise ValueError('invalid s3 filename specified')

    bucket_name, key_name = match.groups()

    if not bucket_name or not key_name:
        raise ValueError('empty bucket and/or key')

    return bucket_name, key_name


def _bucket_exists(bucket_name):
    # http://boto3.readthedocs.io/en/latest/guide/migrations3.html#accessing-a-bucket

    try:
        _s3().meta.client.head_bucket(Bucket=bucket_name)
        return True
    except botocore.exceptions.ClientError as e:
        logger.debug('e.response=%s', e.response)
        if e.response['Error']['Code'] == 'NoSuchBucket':
            return False
        raise  # pragma: no cover


def _object_exists(bucket_name, key_name):

    try:
        _s3().meta.client.head_object(Bucket=bucket_name, Key=key_name)
        return True
    except botocore.exceptions.ClientError as e:
        logger.debug('e.response=%s', e.response)
        if e.response['Error']['Code'] == '404':
            return False
        raise  # pragma: no cover



# -*- coding: utf-8 -*-

"""
Provides useful functions for interacting with the TDI Core API, including creation of
keys, etc.
"""
from __future__ import unicode_literals, division

import os
import base64
import logging

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives.keywrap import aes_key_wrap, aes_key_unwrap

from . import utils

logger = logging.getLogger(__name__)


_BACKEND = default_backend()


def create_aes_key():
    """
    Create an AES256 key for symmetric encryption

    :return: Encryption key bytes
    """
    return os.urandom(32)


def aes_encrypt(plaintext, aes_key, legacy_support=True):
    """
    Encrypt using AES-GCM

    :param plaintext: (string or bytes) that you want encrypted
    :param aes_key: symmetric key to encrypt attribute value with
    :return: Dictionary (Flattened JWE) with base64-encoded ciphertext and base64-encoded iv
    """
    iv = os.urandom(12)
    cipher_alg = Cipher(algorithms.AES(aes_key), modes.GCM(iv), backend=_BACKEND)
    encryptor = cipher_alg.encryptor()
    encr_value = encryptor.update(utils.to_bytes(plaintext)) + encryptor.finalize()
    ciphertext_b64 = utils.base64url_encode(encr_value)
    tag_b64 = utils.base64url_encode(encryptor.tag)
    iv_b64 = base64.b64encode(iv) if legacy_support else utils.base64url_encode(iv)
    ret = {
      "header": {
        "alg": "dir",
        "enc": "A256GCM"
      },
      "iv": iv_b64,
      "ciphertext": ciphertext_b64,
      "tag": tag_b64,
    }

    if legacy_support:
        ct_b64 = base64.b64encode(encr_value + encryptor.tag)
        ret.update({
          "cipher": "aes",
          "mode": "gcm",
          "ts": 128,
          "ct": ct_b64,
        })
    return ret


def aes_decrypt(attr_ct, aes_key):
    """
    Convenience method to decrypt attribute properties

    :param attr_ct: Dictionary (may be a Flattened JWE) with base64-encoded
        ciphertext and base64-encoded iv
    :param aes_key: symmetric key to decrypt attribute value with
    :return: plaintext bytes
    """
    if not isinstance(attr_ct, dict) or (
        attr_ct.get('cipher', 'aes') != 'aes' or
        attr_ct.get('mode', 'gcm') != 'gcm' or
        'iv' not in attr_ct or
        (
            'header' in attr_ct and (
                attr_ct['header'].get('alg', 'dir') != 'dir' or
                attr_ct['header'].get('env', 'A256GCM') != 'A256GCM'
            )
        )
    ):
        raise ValueError('invalid encrypted attribute')

    iv = None
    ciphertext = None

    if 'ciphertext' in attr_ct:
        # JWE included, prefer that
        ciphertext = utils.base64url_decode(attr_ct.get('ciphertext'))
        tag = utils.base64url_decode(attr_ct.get('tag'))

        if len(tag) != 16:  # 128 // 8
            raise ValueError('invalid tag size: {}'.format(len(tag)))

        if 'cipher' in attr_ct:
            # hybrid mode, iv is in legacy format
            iv = base64.b64decode(attr_ct['iv'])
        else:
            iv = utils.base64url_decode(attr_ct['iv'])
    else:
        # legacy only
        ts = attr_ct.get('ts', 128)

        if ts != 128:
            raise ValueError('invalid tag size: {}'.format(ts))

        iv = base64.b64decode(attr_ct['iv'])
        tag_ct = base64.b64decode(attr_ct['ct'])

        sp = ts // 8
        ciphertext = tag_ct[:-sp]
        tag = tag_ct[-sp:]

    cipher_alg = Cipher(
        algorithms.AES(aes_key),
        modes.GCM(iv, tag, min_tag_length=8),
        backend=_BACKEND
    )
    decryptor = cipher_alg.decryptor()
    return decryptor.update(ciphertext) + decryptor.finalize()


def key_wrap(wrapping_key, key_to_wrap):
    return aes_key_wrap(wrapping_key, key_to_wrap, _BACKEND)


def key_unwrap(wrapping_key, wrapped_key):
    return aes_key_unwrap(wrapping_key, wrapped_key, _BACKEND)




from . import plain_file_handler
from . import s3_handler


def join_paths(root, *paths):
    return _get_handler(root).join_paths(root, *paths)


def file_exists(path):
    return _get_handler(path).file_exists(path)


def file_directory_exists(filename):
    return _get_handler(filename).file_directory_exists(filename)


def prepare_directory(dirname):
    return _get_handler(dirname).prepare_directory(dirname)


def prepare_file_directory(filename):
    return _get_handler(filename).prepare_file_directory(filename)


def read_file(filename, binary=True):
    return _get_handler(filename).read_file(filename, binary)


def write_file(filename, data, binary=True):
    return _get_handler(filename).write_file(filename, data, binary)


def _get_handler(path):
    if path.startswith('s3://'):
        return s3_handler
    return plain_file_handler



# -*- coding: utf-8 -*-

"""
Provides useful functions for dealing with JWEs

Based on the
`JSON Web Encryption (JWE) <https://tools.ietf.org/html/rfc7516/>`_, and
`JSON Web Algorithms (JWA) <https://tools.ietf.org/html/rfc7518/>`_,
IETF RFCs.

"""
from __future__ import unicode_literals

import collections
import json
import base64
import logging

from . import jose, service, keychain, utils, exceptions

logger = logging.getLogger(__name__)

NON_OVERRIDABLE_CLAIMS = ['enc', 'alg', 'epk', 'apu']


def make_jwe(raw_claims, sender_keypair, recipient_keypairs, jsonify=True, json_encoder=json.dumps):
    """
    Convert claims into a JWE with General JWE JSON Serialization syntax

    :param raw_claims: payload data that will be converted to json
    :type raw_claims: dict
    :param recipient_keypairs: :py:class:`~oneid.keychain.Keypair`\s to encrypt the claims for
    :type recipient_keypairs: list or :py:class:`~oneid.keychain.Keypair`
    :param jsonify: If True (default), return JSON, otherwise keep as dict
    :type jsonify: bool
    :param json_encoder: encodes a :py:class:`dict` into JSON. Defaults to `json.dumps`
    :type json_encoder: function
    :return: JWE
    :return_type: str or dict
    """

    if not sender_keypair.identity:
        raise exceptions.IdentityRequired

    if not isinstance(recipient_keypairs, collections.Iterable):
        recipient_keypairs = [recipient_keypairs]

    claims = jose.normalize_claims(raw_claims)
    plaintext = json_encoder(claims)
    nonce = claims['jti']

    ephemeral_keypair = service.create_secret_key()
    common_header = _make_header(claims, sender_keypair.identity, ephemeral_keypair)
    cek = service.create_aes_key()
    encrypted_claims = service.encrypt_attr_value(plaintext, cek, legacy_support=False)

    ret = {
        "unprotected": common_header,
        "iv": utils.to_string(encrypted_claims['iv']),
        "ciphertext": utils.to_string(encrypted_claims['ciphertext']),
        "tag": utils.to_string(encrypted_claims['tag']),
        "recipients": [
            _encrypt_to_recipient(cek, sender_keypair.identity, ephemeral_keypair, keypair, nonce)
            for keypair in recipient_keypairs
        ],
    }

    del cek, ephemeral_keypair  # this wont wipe the memory, but is the best we can do in Python

    if jsonify:
        ret = json_encoder(ret)

    return ret


def decrypt_jwe(jwe, recipient_keypair, json_decoder=json.loads):
    """
    Decrypt the claims in a JWE for a given recipient

    :param jwe: JWE to verify and convert
    :type jwe: str
    :param recipient_keypair: :py:class:`~oneid.keychain.Keypair` to use to decrypt.
    :type recipient_keypair: :py:class:`~oneid.keychain.Keypair`
    :param json_encoder: a function to encode a :py:class:`dict` into JSON. Defaults to `json.dumps`
    :returns: claims
    :rtype: dict
    :raises: :py:class:`~oneid.exceptions.InvalidFormatError`: if not a valid JWE
    """

    if not recipient_keypair.identity:
        raise exceptions.IdentityRequired

    jwe_dict = jose.as_dict(jwe, json_decoder)

    if not jose.is_jwe(jwe_dict):
        logger.debug('attempt to decrypt a non-jwe (type: %s): %s', type(jwe), jwe)
        raise exceptions.InvalidFormatError

    shared_header = jose.get_jwe_shared_header(jwe, json_decoder=json.loads)

    recipient = next((
        recipient for recipient in jwe_dict['recipients']
        if recipient['header']['kid'] == str(recipient_keypair.identity)
    ), None)

    if not recipient:
        logger.warning('attempt to decrypt for invalid recipient (%s)', recipient_keypair.identity)
        raise exceptions.InvalidRecipient

    jwk = shared_header.get('epk')
    sender_keypair = keychain.Keypair.from_jwk(jwk)
    apu = utils.base64url_decode(shared_header.get('apu'))
    apv = utils.base64url_decode(recipient['header']['apv'])

    kek = recipient_keypair.ecdh(sender_keypair, party_u_info=apu, party_v_info=apv)
    e_cek = utils.base64url_decode(recipient['encrypted_key'])

    try:
        cek = service.key_unwrap(kek, e_cek)
    except:
        logger.warning('invalid attempt to decrypt CEK, id=%s', recipient_keypair.identity)
        raise exceptions.DecryptionFailed

    # recast into form expected by decrypt_attr_value
    # TODO: have it be more flexible?
    #
    attr_value = {
        key: base64.b64encode(utils.base64url_decode(value))
        for key, value in jwe_dict.items()
        if key in ['iv', 'ciphertext', 'tag']
    }
    # 'cipher': 'aes', mode': 'gcm' are assumed, 'ts': 128 isn't needed

    try:
        claims = service.decrypt_attr_value(attr_value, cek)
    except:
        logger.warning('invalid attempt to decrypt claims, id=%s', recipient_keypair.identity)
        raise exceptions.DecryptionFailed

    return json_decoder(utils.to_string(claims))


def _make_header(claims, sender_identity, ephemeral_keypair):
    HEADER_EXTRACTED_CLAIMS = ['iss', 'jti', 'nbf', 'exp', 'aud']

    common = {k: v for k, v in claims.items() if k in HEADER_EXTRACTED_CLAIMS}
    nonce = claims['jti']
    epk = ephemeral_keypair.jwk_public

    if any([k in claims for k in NON_OVERRIDABLE_CLAIMS]):
        raise ValueError

    common.update({
        "enc": "A256GCM",
        "alg": "ECDH-ES+A256KW",
        "epk": epk,
        "apu": utils.to_string(utils.base64url_encode(str(sender_identity) + nonce)),
    })
    return common


def _encrypt_to_recipient(cek, sender_identity, ephemeral_keypair, recipient_keypair, nonce):
    apu = utils.to_bytes(str(sender_identity) + nonce)
    apv = utils.to_bytes(str(recipient_keypair.identity) + nonce)

    ret = {
      "header": {
        "kid": recipient_keypair.identity,
        "apv": utils.to_string(utils.base64url_encode(apv)),
      },
    }
    kek = ephemeral_keypair.ecdh(recipient_keypair, party_u_info=apu, party_v_info=apv)
    encrypted_key = service.key_wrap(kek, cek)
    ret['encrypted_key'] = utils.to_string(utils.base64url_encode(encrypted_key))

    del kek

    return ret



from __future__ import unicode_literals

"""
Provides functionality to verify signatures from oneID users,
specifically as part of the Javascript API `signin()`_ callback.

.. _signin(): https://developer.oneid.com/docs/#/login/javascript-api/signin
"""
import copy
import json
import requests


class OneIDAuthenticationService:
    """
    Encapsulates a connection to the oneID servers

    :param api_id: Your OneID API ID credentials (from https://keychain.oneid.com/register)
    :param api_key: Your OneID API Key credentials (from https://keychain.oneid.com/register)
    :param server_flag: If you want to connect to a different API  should be (for example)
                        "-test" when using a non-production server
    """
    def __init__(self, api_id=None, api_key=None, server_flag=""):
        self.keychain_server = "https://keychain%s.oneid.com" % server_flag

        # Set the API credentials
        self.set_credentials(api_id, api_key)

    def _call_keychain(self, method, data={}):
        """
        Call the OneID Keychain Service. (i.e. to validate signatures)

        :param method: The OneID API call you wish to call
        :param data: Data for the OneID API CAll
        """
        url = "%s/%s" % (self.keychain_server, method)
        r = requests.post(url, json.dumps(data), auth=(self.api_id, self.api_key))
        return r.json()

    def set_credentials(self, api_id="", api_key=""):
        """
        Set the credentials used for access to the OneID Helper Service

        :param api_id: Your OneID API ID
        :param api_key: Your OneID API key
        """
        if not api_id or not api_key:
            raise ValueError('api_id and api_key are required')

        self.api_id = api_id
        self.api_key = api_key

    def validate(self, oneid_payload):
        """
        Validate the data received by a callback

        :param oneid_payload: The dictionary you want to validate,
            typically the payload from a OneID sign in call
        :return: if successful, `oneid_payload`, updated with the response from oneID.
            Otherwise, the error response from oneID.
        """
        if isinstance(oneid_payload, dict):
            oneid_payload = copy.deepcopy(oneid_payload)
        else:
            oneid_payload = json.loads(oneid_payload)

        data_to_validate = {
            "nonces": oneid_payload["nonces"],
            "uid": oneid_payload["uid"]
        }

        if "attr_claim_tokens" in oneid_payload:
            data_to_validate["attr_claim_tokens"] = oneid_payload["attr_claim_tokens"]

        keychain_response = self._call_keychain("validate", data_to_validate)

        if not self.success(keychain_response):
            keychain_response["failed"] = "failed"
            return keychain_response

        oneid_payload.update(keychain_response)

        return oneid_payload

    def success(self, oneid_response):
        """
        Check errorcode in a response

        :param oneid_response: A response from :py:meth:`validate()`
        :return: True if the response indicates success, False otherwise.
        """
        return oneid_response.get("errorcode", -1) == 0



from __future__ import unicode_literals

import os
import yaml
import collections
import logging

from requests import request
from codecs import open

from . import service, jose, jwts, jwes, exceptions

logger = logging.getLogger(__name__)


class SessionBase(object):
    """
    Abstract Session Class

    :ivar identity_credentials: TDI identity :class:`~oneid.keychain.Credentials`
    :ivar project_credentials: unique project credentials :class:`~oneid.keychain.Credentials`
    :ivar oneid_credentials: TDI project credentials :class:`~oneid.keychain.Credentials`
    :ivar oneid_credentials: peer credentials :class:`~oneid.keychain.Credentials`
    :ivar config: Dictionary or configuration keyword arguments
    """
    def __init__(self, identity_credentials=None, project_credentials=None,
                 oneid_credentials=None, peer_credentials=None, config=None):
        """

        :param identity_credentials: :py:class:`~oneid.keychain.Credentials`
        :param project_credentials: :py:class:`~oneid.keychain.ProjectCredentials`
        :param oneid_credentials: :py:class:`~oneid.keychain.Credentials`
        :param peer_credentials: list of :py:class:`~oneid.keychain.Credentials`
            If provided, session will be encrypted to recipients
        :param config: Dictionary or configuration keyword arguments
        :return:
        """
        self.identity_credentials = identity_credentials
        self.project_credentials = project_credentials
        self.oneid_credentials = oneid_credentials
        self.peer_credentials = peer_credentials
        if peer_credentials and not isinstance(peer_credentials, collections.Iterable):
            self.peer_credentials = [peer_credentials]

    def _load_config(self, config_file):
        """
        Load configuration from file
        :return: dict()
        """
        # Load params from configuration file
        with open(config_file, mode='r', encoding='utf-8') as config:
            params = yaml.safe_load(config)
            return params

    def _create_services(self, methods, **kwargs):
        """
        Populate session variables and create methods from args

        :return: None
        """
        service_creator = service.ServiceCreator()

        for method in methods:
            if method != 'GLOBAL':
                setattr(self, method,
                        service_creator.create_service_class(method,
                                                             methods[method],
                                                             self,
                                                             **kwargs)
                        )

    def _get_recipient_keypairs(self, encrypt_to_peers, other_recipients):
        recipients = []

        if encrypt_to_peers and self.peer_credentials:
            recipients += self.peer_credentials

        if other_recipients:
            recipients += other_recipients

        return [c.keypair for c in recipients]

    def make_http_request(self, http_method, url, headers=None, body=None):
        """
        Generic HTTP request

        :param headers:
        :param body:
        :return:
        """
        valid_http_methods = ['GET', 'PUT', 'POST', 'DELETE']
        if http_method not in valid_http_methods:
            raise TypeError('HTTP method must be %s' %
                            ', '.join(valid_http_methods))

        req = request(http_method, url, headers=headers, data=body)

        logger.debug(
            'making http %s request to %s, headers=%s, data=%s, req=%s',
            http_method, url, headers, body, req,
        )

        # 403 is Forbidden, raise an error if this occurs
        if req.status_code == 403:
            raise exceptions.InvalidAuthentication()

        return req.content

    def service_request(self, http_method, endpoint, body=None):
        """
        Make an API Request

        :param method:
        :param endpoint:
        :param body:
        :return:
        """
        auth_jwt_header = jwts.make_jwt({}, self.identity_credentials.keypair)

        headers = {
            'Content-Type': 'application/jwt',
            'Authorization': 'Bearer %s' % auth_jwt_header
        }
        return self.make_http_request(http_method, endpoint, headers=headers, body=body)

    def prepare_message(self, *args, **kwargs):
        raise NotImplementedError

    def send_message(self, *args, **kwargs):
        raise NotImplementedError

    def verify_message(self, *args, **kwargs):
        raise NotImplementedError


class DeviceSession(SessionBase):
    def __init__(self, identity_credentials=None, project_credentials=None,
                 oneid_credentials=None, peer_credentials=None, config=None):
        super(DeviceSession, self).__init__(identity_credentials,
                                            project_credentials,
                                            oneid_credentials,
                                            peer_credentials, config)

    def verify_message(self, message, rekey_credentials=None):
        """
        Verify a message received from the Project

        :param message: JSON formatted JWS with at least two signatures
        :param rekey_credentials: List of :class:`~oneid.keychain.Credential`
        :return: verified message or False if not valid
        """
        standard_keypairs = [
            self.project_credentials.keypair,
            self.oneid_credentials.keypair,
        ]

        if rekey_credentials:
            keypairs = [credentials.keypair for credentials in rekey_credentials]

            kids = [sig_params['kid'] for sig_params in jwts.get_jws_key_ids(message)]
            keypairs += [keypair for keypair in standard_keypairs if keypair.identity in kids]
        else:
            keypairs = standard_keypairs

        ret = jwts.verify_jws(message, keypairs)

        if jose.is_jwe(ret):
            ret = jwes.decrypt_jwe(ret, self.identity_credentials.keypair)

        return ret

    def prepare_message(self, encrypt_to_peers=True, other_recipients=None, *args, **kwargs):
        """
        Prepare a message before sending

        :param encrypt_to_peers: If True (default), and peer_credentials available,
            encrypt the message to them
        :type encrypt_to_peers: bool
        :param other_recipients: Additional recipients to encrypt to
        :type other_recipients: list of :class:`~oneid.keychain.Credential`
        :return: Signed JWS
        """
        claims = kwargs
        claims['iss'] = self.identity_credentials.id

        recipient_keypairs = self._get_recipient_keypairs(encrypt_to_peers, other_recipients)

        if recipient_keypairs:
            claims = jwes.make_jwe(
                claims, self.identity_credentials.keypair, recipient_keypairs, jsonify=False
            )

        return jwts.make_jws(claims, self.identity_credentials.keypair)

    def add_signature(self, message, default_jwt_kid=None):
        """
        Add this Device's signature to a message

        Note that the semantics of this signature are application-specific. If the
        application expects only verified messages to be co-signed, the caller is
        responsible for verifying the message first. Otherwise, the signature only
        indicates that the message was processed by this Device.

        Likewise, this method will not decrypt a JWE. If the message was encrypted
        for this Device, and should be decrypted and re-signed, the caller should
        do that through other means, such as :func:`verify_message` and :func:`prepare_message`.

        :param str message: Previously-signed JWS (Compact or JSON) or JWT
        :param str default_jwt_kid: (optional) value for 'kid' header field if passing a JWT
            without one
        :return: Signed JWS with additional Device signature
        """
        return jwts.extend_jws_signatures(
            message, self.identity_credentials.keypair, default_jwt_kid
        )

    def send_message(self, *args, **kwargs):
        raise NotImplementedError


class ServerSession(SessionBase):
    """
    Enable Server to request two-factor Authentication from TDI Core
    """
    def __init__(self, identity_credentials=None, project_credentials=None,
                 oneid_credentials=None, peer_credentials=None, config=None):
        super(ServerSession, self).__init__(identity_credentials,
                                            project_credentials,
                                            oneid_credentials,
                                            peer_credentials, config)

        if isinstance(config, dict):
            params = config
        else:
            # Load default
            default_config = os.path.join(os.path.dirname(__file__),
                                          'data', 'oneid_server.yaml')
            params = self._load_config(config if config else default_config)

        self._create_services(params)

    def _create_services(self, params, **kwargs):
        """
        Populate session variables and create methods from
        :return: None
        """
        global_kwargs = params.get('GLOBAL', {})
        if self.project_credentials:
            global_kwargs['project_credentials'] = self.project_credentials

        super(ServerSession, self)._create_services(params, **global_kwargs)

    def prepare_message(self, rekey_credentials=None, encrypt_to_peers=True,
                        other_recipients=None, **kwargs):
        """
        Build message that has two-factor signatures

        :param rekey_credentials: (optional) rekey credentials
        :type rekey_credentials: list
        :param encrypt_to_peers: If True (default), and peer_credentials available,
            encrypt the message to them
        :type encrypt_to_peers: bool
        :param other_recipients: Additional recipients to encrypt to
        :type other_recipients: list of :class:`~oneid.keychain.Credential`
        :return: Signed JWS to be sent to devices
        """
        if self.project_credentials is None:
            raise AttributeError

        keypairs = [
            self.project_credentials.keypair
        ]

        if rekey_credentials:
            keypairs += [credentials.keypair for credentials in rekey_credentials]

        claims = kwargs
        claims['iss'] = self.identity_credentials.id

        recipient_keypairs = self._get_recipient_keypairs(encrypt_to_peers, other_recipients)

        if recipient_keypairs:
            claims = jwes.make_jwe(
                claims, self.identity_credentials.keypair, recipient_keypairs, jsonify=False
            )

        jws = jwts.make_jws(claims, self.identity_credentials.keypair)

        oneid_response = self.authenticate.server(
            project_id=self.project_credentials.keypair.identity,
            identity=self.identity_credentials.keypair.identity,
            body=jws
        )

        if not oneid_response:
            logger.debug('TDI Core refused to co-sign server message')
            raise exceptions.InvalidAuthentication

        stripped_response = jwts.remove_jws_signatures(
            oneid_response, self.identity_credentials.id
        )
        return jwts.extend_jws_signatures(stripped_response, keypairs)

    def send_message(self, *args, **kwargs):
        raise NotImplementedError

    def verify_message(self, message, device_credentials, get_oneid_cosignature=True):
        """
        Verify a message received from/through one or more Devices

        :param message: JSON formatted JWS or JWT signed by the Device
        :param device_credentials: :class:`~oneid.keychain.Credential` (or list of them)
            to verify Device signature(s) against
        :param get_oneid_cosignature: (default: True) verify with TDI Core first
        :return: verified message or False if not valid
        """

        if not device_credentials:
            raise AttributeError

        if not isinstance(device_credentials, collections.Iterable):
            device_credentials = [device_credentials]

        keypairs = [credential.keypair for credential in device_credentials]

        if get_oneid_cosignature:
            keypairs += [self.oneid_credentials.keypair]

            # TODO: if not already signed by Core: (for now, do as asked, let caller deal with it)

            if len(device_credentials) == 1:
                message = self.authenticate.edge_device(
                    project_id=self.project_credentials.keypair.identity,
                    identity=keypairs[0].identity,
                    body=message,
                )
            else:
                message = self.authenticate.project(
                    project_id=self.project_credentials.keypair.identity,
                    body=message,
                )

            if not message:
                logger.debug('TDI Core refused to co-sign device message')
                raise exceptions.InvalidAuthentication

        ret = jwts.verify_jws(message, keypairs)

        if jose.is_jwe(ret):
            try:
                # the message is most likely intended for the Project
                ret = jwes.decrypt_jwe(ret, self.project_credentials.keypair)
            except exceptions.InvalidRecipient:
                # but in case it is intended for the Project Server itself
                ret = jwes.decrypt_jwe(ret, self.identity_credentials.keypair)

        return ret


class AdminSession(SessionBase):
    """
    Admin Users will only interface with TDI Core service,
    They only need an identity_credentials and oneid_credentials
    to verify responses
    """
    def __init__(self, identity_credentials, project_credentials=None,
                 oneid_credentials=None, config=None):
        super(AdminSession, self).__init__(identity_credentials,
                                           project_credentials,
                                           oneid_credentials, config)

        if isinstance(config, dict):
            params = config
        else:
            default_config = os.path.join(os.path.dirname(__file__),
                                          'data', 'oneid_admin.yaml')
            params = self._load_config(config if config else default_config)

        self._create_services(params)

    def _create_services(self, params, **kwargs):
        """
        Populate session variables and create methods from
        :return: None
        """
        global_kwargs = params.get('GLOBAL', {})
        if self.project_credentials:
            global_kwargs['project_credentials'] = self.project_credentials

        super(AdminSession, self)._create_services(params, **global_kwargs)

    def prepare_message(self, *args, **kwargs):
        raise NotImplementedError

    def send_message(self, *args, **kwargs):
        raise NotImplementedError

    def verify_message(self, *args, **kwargs):
        raise NotImplementedError



"""
Helpful utility functions
"""
from __future__ import unicode_literals
import six

import base64
import logging

from datetime import datetime
from dateutil import parser, tz

logger = logging.getLogger(__name__)

UTC = tz.tzutc()
EPOCH = datetime(1970, 1, 1, tzinfo=UTC)


def to_bytes(data):
    return data.encode('utf-8') if isinstance(data, six.text_type) else data


def to_string(data):
    return data if isinstance(data, six.text_type) else data.decode('utf-8')


def to_timestamp(dt):
    if not isinstance(dt, datetime):
        dt = parser.parse(dt)
    return (dt - EPOCH).total_seconds()


def from_timestamp(timestamp):
    return datetime.fromtimestamp(timestamp, UTC)


def base64url_encode(msg):
    """
    Default b64_encode adds padding, jwt spec removes padding
    :param input:
    :type input: string or bytes
    :return: base64 en
    :rtype: bytes
    """
    encoded_input = base64.urlsafe_b64encode(to_bytes(msg))
    stripped_input = to_bytes(to_string(encoded_input).replace('=', ''))
    return stripped_input


def base64url_decode(msg):
    """
    JWT spec doesn't allow padding characters. base64url_encode removes them,
    base64url_decode, adds them back in before trying to base64 decode the message

    :param msg: URL safe base64 message
    :type msg: string or bytes
    :return: decoded data
    :rtype: bytes
    """
    bmsg = to_bytes(msg)
    pad = len(bmsg) % 4
    if pad > 0:
        bmsg += b'=' * (4 - pad)

    return base64.urlsafe_b64decode(bmsg)



from __future__ import unicode_literals
__all__ = [
    "__title__", "__summary__", "__uri__", "__version__", "__author__",
    "__email__", "__license__", "__copyright__",
]

__title__ = "oneID-connect"
__summary__ = ("Neustar TDI is an identity & authentication "
               "framework for the Internet of Things (IoT) "
               "devices, servers, and end-users.")

__uri__ = "https://github.com/OneID/oneID-connect-python"

__version__ = "0.20.0"

__author__ = "Neustar Inc."
__email__ = "support@oneID.com"

__license__ = "Apache License, Version 2.0"
__copyright__ = "Copyright 2012-2016 {0}".format(__author__)



# from __future__ import unicode_literals
import os

from setuptools import setup, find_packages
from codecs import open

base_dir = os.path.dirname(__file__)
src_dir = os.path.join(base_dir, "src")


about = {}
with open(os.path.join(src_dir, "oneid", "__about__.py"), encoding='utf-8') as f:
    exec(f.read(), about)

with open(os.path.join(base_dir, "README.rst"), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name=about['__title__'],
    version=about['__version__'],
    description=about['__summary__'],
    long_description=long_description,
    url=about['__uri__'],
    author=about['__author__'],
    author_email=about['__email__'],
    license=about['__license__'],
    classifiers=[
        'Development Status :: 4 - Beta',
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX",
        "Operating System :: POSIX :: BSD",
        "Operating System :: POSIX :: Linux",
        'Topic :: Security :: Cryptography',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],

    keywords='Neustar Trusted Device Identity (TDI) IoT Authentication',
    package_dir={"": "src"},
    packages=find_packages(where='src',
                           exclude=['contrib', 'docs', 'tests*',
                                    'venv', 'example*', '*egg-info',
                                    '.gitignore']),
    package_data={
        'oneid': ['data/*.yaml'],
    },
    install_requires=['cryptography>=2.0.2,<2.1', 'PyYAML>=3.11,<4',
                      'requests[security]>=2.9.1,<2.10', 'python-dateutil>=2.4.2,<2.5',
                      'pytz>=2017.2', 'six>=1.10.0,<1.11', 'boto3>=1.4.4,<1.5'],
)
